from django_filters.rest_framework import DjangoFilterBackend
from global_functions.utils import get_prop
from rest_framework import filters, status, viewsets
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response


class NovadataModelViewSet(viewsets.ModelViewSet):
    permission_classes = [
        IsAuthenticated,
    ]

    filter_backends = [
        filters.SearchFilter,
        DjangoFilterBackend,
        filters.OrderingFilter,
    ]

    filterset_fields: list = None

    ordering_fields: list = None

    search_fields: list = None

    auto_search_fields: bool = False

    relation_fields = [
        "OneToOneField",
        "ForeignKey",
    ]

    def get_filterset_fields(self):
        """Retorna os campos de filtro."""
        model = self.serializer_class().Meta.model
        self.filterset_fields = get_prop(
            model,
            "filterset_fields",
            str=False,
        )

        return self.filterset_fields

    def get_ordering_fields(self):
        """Retorna os campos de ordenação."""
        model = self.serializer_class().Meta.model
        self.ordering_fields = get_prop(
            model,
            "ordering_fields",
            str=False,
        )

        return self.ordering_fields

    def get_search_fields(self):
        """Retorna os campos de busca."""
        model = self.serializer_class().Meta.model
        self.search_fields = get_prop(
            model,
            "search_fields",
            str=False,
        )

        return self.search_fields

    def get_fk_fields(self):
        """Retorna os campos de relacionamento."""
        model = self.serializer_class().Meta.model
        fields = model._meta.get_fields()
        fk_fields = [
            (field.name, field.remote_field.model)
            for field in fields
            if field.get_internal_type() in self.relation_fields
        ]

        return fk_fields

    def get_data(self, request):
        """Retorna os dados da requisição de forma tratada."""
        data = request.data.copy()
        for fk_field in self.get_fk_fields():
            field_name = fk_field[0]
            try:
                if isinstance(data[field_name], dict):
                    data[field_name] = data[field_name].get(
                        "id",
                        None,
                    )
            except KeyError:
                print(f"Não passou {field_name}")

        return data

    def create(self, request, *args, **kwargs):
        """Realiza o create de um objeto."""
        serializer = self.get_serializer(data=self.get_data(request))
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(
            serializer.data, status=status.HTTP_201_CREATED, headers=headers
        )

    def update(self, request, *args, **kwargs):
        """Realiza o update de um objeto."""
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        data = self.get_data(request)

        obj_datas = {}
        for field in self.get_fk_fields():
            field_name = field[0]
            obj_datas[field_name] = data.pop(
                field_name,
                None,
            )

        serializer = self.get_serializer(
            instance,
            data=data,
            partial=partial,
        )
        serializer.is_valid(raise_exception=True)

        for field in self.get_fk_fields():
            field_name = field[0]

            try:
                model = field[1]
                obj_data = obj_datas[field_name]
                if obj_data:
                    if isinstance(obj_data, dict):
                        fk_instance = model.objects.get(
                            pk=obj_data.get("id", None)
                        )
                    elif isinstance(obj_data, list):
                        fk_instance = model.objects.get(pk=obj_data[0])
                    else:
                        fk_instance = model.objects.get(pk=obj_data)

                    setattr(serializer.instance, field_name, fk_instance)
            except model.DoesNotExist:
                print(f"Não passou {field_name}")

        self.perform_update(serializer)
        serializer.instance.save()

        return Response(serializer.data)

    def __init__(self, **kwargs) -> None:
        """Método para executarmos ações ao iniciar a classe."""
        super().__init__(**kwargs)
        if not self.filterset_fields:
            self.get_filterset_fields()

        if not self.ordering_fields:
            self.get_ordering_fields()

        if self.auto_search_fields and not self.search_fields:
            self.get_search_fields()



from collections import namedtuple


def get_changes(object):
    """
    Função para pegar as mudanças ocorridas em um objeto.

    Deve ser chamada dentro de uma função save.
    Parâmetros:
        object: objeto que você deseja saber as mudanças.
    Retorno:
        namedtuple: com os campos:
            normais: Lista com os nomes dos campos que mudaram.
            back: Lista com os objetos dos campos que mudaram.
            old_and_new_object: Dicionário com os campos:
                old_object: Objeto antes do save.
                new_object: Objeto durante o save.
    """
    object_class = object.__class__
    old_object = object_class.objects.get(pk=object.pk)
    new_object = object
    mudancas = namedtuple(
        "mudancas", ["normais", "back", "old_and_new_object"]
    )

    def houve_mudanca_normal(field):
        """Retorna o nome do campo se houve mudança e None se não houve."""
        try:
            if getattr(old_object, field.name) != getattr(
                new_object, field.name
            ):
                return field.name
        except AttributeError:
            ...

    def houve_mudanca_back(field):
        try:
            if getattr(old_object, field.name) != getattr(
                new_object, field.name
            ):
                return field
        except AttributeError:
            ...

    change_normal_fields = list(
        filter(
            campo_valido,
            map(
                houve_mudanca_normal,
                object._meta.fields,
            ),
        )
    )

    change_back_fields = list(
        filter(
            campo_valido,
            map(
                houve_mudanca_back,
                object._meta.get_fields(),
            ),
        )
    )
    old_and_new_object = {"old_object": old_object, "new_object": new_object}

    return mudancas(
        change_normal_fields, change_back_fields, old_and_new_object
    )


def campo_valido(field):
    """Retorna True se o campo for válido e False se não for."""
    return True if field else False



from pyclbr import Class

from crum import get_current_user


def create_logs(
    parent_object: object,
    back_update_fields: list,
    old_and_new_object: dict,
    log_class: Class,
    fk_name: str,
) -> list:
    """
    Função para criar objetos log para diferentes tipos de objetos pais.

    Parâmetros:
        parent_object: Objeto pai do log.
        back_update_fields: Lista de campos que foram alterados.
        old_and_new_object: Dicionário com os valores:
            old_object: Objeto antes do save.
            new_object: Objeto durante o save.
        log_class: Classe do log.
        fk_name: Nome do campo FK que está apontando para o pai no log.
    Retorna:
        Lista de objetos log criados.
    """

    def create_log_instance(field):
        """Cria uma instância de log para cada campo alterado."""
        valor_anterior = get_value(old_and_new_object["old_object"], field)
        valor_posterior = get_value(old_and_new_object["new_object"], field)

        instance = log_class(
            usuario=get_current_user(),
            campo_back=field.name,
            campo_front=field.verbose_name,
            valor_anterior=valor_anterior,
            valor_posterior=valor_posterior,
        )

        setattr(instance, fk_name, parent_object)
        return instance.save()

    return list(map(create_log_instance, back_update_fields))


def get_value(object, field):
    """Retorna o valor do campo do objeto."""
    display_name = f"get_{field.name}_display"
    has_display_name = hasattr(object, display_name)

    if has_display_name:
        return getattr(object, display_name)()
    else:
        return getattr(object, field.name)



from rest_framework import serializers
from rest_framework.fields import empty


class NovadataModelSerializer(serializers.ModelSerializer):
    representation_fields: list = []

    def to_representation(self, instance):
        if instance and not isinstance(instance, list):
            default_return = super(
                NovadataModelSerializer, self
            ).to_representation(instance)

            if self.representation_fields:
                for field_name, serializer in self.representation_fields:
                    default_return[field_name] = serializer(
                        getattr(instance, field_name)
                    ).data

            return default_return

    def __init__(self, instance=None, data=empty, **kwargs):
        super(NovadataModelSerializer, self).__init__(instance, data, **kwargs)
        self.to_representation(instance)



from advanced_filters.admin import AdminAdvancedFiltersMixin
from django.contrib import admin
from import_export.admin import ImportExportModelAdmin
from novadata_utils.functions import get_prop


class NovadataModelAdmin(
    ImportExportModelAdmin,
    AdminAdvancedFiltersMixin,
    admin.ModelAdmin,
):
    """
    Classe para realizar funcionalidades default em todas as classes do admin.

    A mesma adiciona todos os campos possíveis nas seguintes propriedades:
    - list_display
    - list_filter
    - autocomplete_fields
    - filter_horizontal
    """

    list_display: list = []

    search_fields: list = []

    list_filter: list = []

    autocomplete_fields: list = []

    auto_search_fields: bool = False

    filter_horizontal: list = []

    def get_list_display(self, request):
        """Retorna a lista de campos que estarão na listagem."""
        super().get_list_display(request)

        if not self.list_display:
            model = self.model
            list_display = get_prop(model, "list_display")

            return list_display
        else:
            return self.list_display

    def get_search_fields(self, request):
        """Retorna a lista de campos que estarão no campo de busca."""
        super().get_search_fields(request)

        if self.auto_search_fields and not self.search_fields:
            model = self.model
            search_fields = get_prop(model, "search_fields")

            return search_fields
        else:
            return self.search_fields

    def get_list_filter(self, request):
        """Retorna a lista de campos que estarão no filtro."""
        super().get_list_filter(request)

        if not self.list_filter:
            model = self.model
            list_filter = get_prop(model, "list_filter")

            return list_filter
        else:
            return self.list_filter

    def get_autocomplete_fields(self, request):
        """Retorna a lista de campos que estarão no autocomplete."""
        super().get_autocomplete_fields(request)

        if not self.autocomplete_fields:
            model = self.model
            autocomplete_fields = get_prop(model, "autocomplete_fields")

            return autocomplete_fields
        else:
            return self.autocomplete_fields

    def get_filter_horizontal(self):
        """Retorna a lista de campos que estarão no filtro horizontal."""
        if not self.filter_horizontal:
            model = self.model
            filter_horizontal = get_prop(model, "filter_horizontal")

            return filter_horizontal
        else:
            return self.filter_horizontal

    def __init__(self, *args, **kwargs):
        """Método para executarmos ações ao iniciar a classe."""
        super().__init__(*args, **kwargs)
        self.filter_horizontal = self.get_filter_horizontal()



from django.contrib.auth.models import User
from django.db import models


class NovadataModel(models.Model):
    data_criacao = models.DateTimeField(
        verbose_name="Data de criação",
        auto_now_add=True,
    )

    data_atualizacao = models.DateTimeField(
        verbose_name="Data de atualização",
        auto_now=True,
    )

    usuario_criacao = models.ForeignKey(
        User,
        on_delete=models.SET_NULL,
        verbose_name="Usuário de criação",
        blank=True,
        null=True,
    )

    usuario_atualizacao = models.ForeignKey(
        User,
        related_name="%(class)s_requests_modified",
        verbose_name="Usuário de atualização",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    def save(self, *args, **kwargs):
        """Sobrescrita do método save para realizarmos ações personalizadas."""
        from crum import get_current_user

        user = get_current_user()
        if user and not user.pk:
            user = None
        if not self.pk:
            self.usuario_criacao = user
        self.usuario_atualizacao = user

        super(NovadataModel, self).save(*args, **kwargs)



from django.apps import AppConfig


class NovadataUtilsConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'novadata_utils'

    def ready(self):
        import novadata_utils.signals



from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.http import urlencode


def reverse_lazy_plus(
    url_name: str,
    url_params: list = [],
    get_params: dict = {},
    hash: str = "",
    just_uri: bool = False,
) -> HttpResponseRedirect:
    """
    Uma reverse_lazy 2.0.

    Aceita parâmetros GET e alguma #, Além de parâmetros de url.

    Caso desejse somente a url, ou seja, somente a string gerada
    (útil em funções como get_success_url), passe o último parâmetro
    (just_uri) como True.

    Usage::

        reverse_lazy_plus(
            'produto',
            url_params=[1, 'pedido_lente_contato'],
            get_params={'mensagem': 'Esta é uma mensagem'},
            '#aba-6',
        )
        # Output:
        # /produto/1/pedido_lente_contato?mensagem=Esta%20é%20uma%20mensagem#aba-6
    """
    url = reverse(url_name, args=url_params)
    params = urlencode(get_params).replace("+", " ")
    params = f"?{params}" if params else ""

    final_url = f"{url}{params}{hash}"
    return final_url if just_uri else HttpResponseRedirect(final_url)



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.db.models import Q


class LoginUsernameEmail(ModelBackend):
    def authenticate(self, request, **kwargs):
        """Função para autenticar o usuário."""
        try:
            user = (
                get_user_model()
                .objects.filter(
                    Q(username=kwargs["username"])
                    | Q(email=kwargs["username"])
                )
                .first()
            )
            return (
                user
                if user and user.check_password(kwargs["password"])
                else None
            )
        except KeyError:
            return None



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    readme = fh.read()

setup(
    name="novadata_utils",
    version="0.1.4",
    url="https://github.com/TimeNovaData/novadata_utils/",
    license="MIT License",
    author="Flávio Silva",
    long_description=readme,
    long_description_content_type="text/markdown",
    author_email="flavio.nogueira.profissional@gmail.com",
    keywords="Django, utils, ndt, novadata, nova data, nova, data",
    description="novadata utils",
    packages=find_packages(),
    install_requires=[
        "django",
        "djangorestframework",
        "django-advanced-filters",
        "django-import-export",
    ],
    project_urls={
        "GitHub": "https://github.com/TimeNovaData/novadata_utils/",
    },
)



props_dict = {
    # Admin props
    "list_display": [
        "BigAutoField",
        "BooleanField",
        "CharField",
        "DateField",
        "DateTimeField",
        "DecimalField",
        "ForeignKey",
        "IntegerField",
        "PositiveIntegerField",
    ],
    "search_fields": [
        "BigAutoField",
        "CharField",
        "DateField",
        "DateTimeField",
        "DecimalField",
        "IntegerField",
        "PositiveIntegerField",
    ],
    "list_filter": [
        "BooleanField",
        "DateField",
        "DateTimeField",
        "ForeignKey",
    ],
    "autocomplete_fields": [
        "ForeignKey",
    ],
    "filter_horizontal": [
        "ManyToManyField",
    ],
    # Generic props
    "foreign_keys": [
        "ForeignKey",
    ],
    # Viewset props
    "filterset_fields": [
        "BooleanField",
        "DateField",
        "DateTimeField",
        "ForeignKey",
    ],
    "ordering_fields": [
        "BigAutoField",
        "CharField",
        "DateField",
        "DateTimeField",
        "DecimalField",
        "IntegerField",
        "PositiveIntegerField",
        "BooleanField",
        "ForeignKey",
    ],
}


def get_prop(model, prop, str=False):
    """
    Retorna uma lista de campos de um model baseado em uma propriedade.
    Exemplo:
        get_prop(model, "list_display") retorna todos os campos que podem ser
        exibidos na listagem do admin. Que são: 
            "BigAutoField",
            "BooleanField",
            "CharField",
            "DateField",
            "DateTimeField",
            "DecimalField",
            "ForeignKey",
            "IntegerField" e
            "PositiveIntegerField".
    
    """
    
    props = []
    fields = model._meta.get_fields()
    for field in fields:
        field_type = field.get_internal_type()
        is_original_field = not hasattr(field, "field")

        if field_type in props_dict[prop] and is_original_field:
            if str:
                field_str = f'"{field.name}",'
                props.append(field_str)
            else:
                props.append(field.name)

    return props
